/* *INDENT-OFF* */
!/* Fiasco - computes sample statistics.
!   Copyright (C) 1997 Free Software Foundation, Inc.
!
!   This program is free software; you can redistribute it and/or
!   modify it under the terms of the GNU General Public License as
!   published by the Free Software Foundation; either version 2 of the
!   License, or (at your option) any later version.
!
!   This program is distributed in the hope that it will be useful, but
!   WITHOUT ANY WARRANTY; without even the implied warranty of
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
!   General Public License for more details.
!
!   You should have received a copy of the GNU General Public License
!   along with this program; if not, write to the Free Software
!   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!   02111-1307, USA.
!
!   The author may be contacted at <pfaffben@pilot.msu.edu> on the
!   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
!   through more mundane means. */
!
!/*
! LIST procedure.
!*/
!

$ list:
$   *variables=varlist("PV_NO_SCRATCH");
$   cases=:from n:first,"%s>0"/by n:step,"%s>0"/*to n:last,"%s>0";
$   format=numbering:numbered/!unnumbered,
$          wrap:!wrap/single,
$          weight:weight/!noweight.
/* *INDENT-ON* */

/* Main declarations. */

#include <config.h>
#include <assert.h>
#include <stdio.h>
#include "output.h"
#include "som.h"
#include "misc.h"

/* (headers) */

/* Declare virtual local variables. */
#define EXTERN extern
#include "somP.h"
#undef EXTERN

#undef DEBUGGING
/*#define DEBUGGING 1 */
#include "debug-print.h"

#if DEBUGGING
static void debug_print (void);
#endif

/* (decls global) */

/* Layout for one output driver. */
typedef struct
  {
    int type;		/* 0=Values and labels fit across the page. */
    int n_vertical;	/* Number of labels to list vertically. */
    int header_rows;	/* Number of header rows. */
    char **header;	/* The header itself. */
  }
list_ext;

/* Current case number. */
static int case_num;

/* Line buffer. */
static char *line_buf;

/* TTY-style output functions. */
static int n_lines_remaining (outp_driver *d);
static int n_chars_width (outp_driver *d);
static void write_line (outp_driver *d, char *s);

/* Other functions. */
static int list_cases (ccase *);
static void flush_table (int this_part);
static void determine_layout (void);
static void clean_up (void);
static void write_header (outp_driver *);
static void write_all_headers (void);

/* Returns the number of text lines that can fit on the remainder of
   the page. */
static inline int
n_lines_remaining (outp_driver *d)
{
  som_driver_ext *som = d->som;
  int diff;

  diff = d->length - som->cp_y;
  return (diff > 0) ? (diff / som->font_height) : 0;
}

/* Returns the number of fixed-width character that can fit across the
   page. */
static inline int
n_chars_width (outp_driver *d)
{
  som_driver_ext *som = d->som;

  return d->width / som->fixed_width;
}

/* Writes the line S at the current position and advances to the next
   line.  */
static void
write_line (outp_driver *d, char *s)
{
  som_driver_ext *som = d->som;
  outp_text text;
  
  assert (som->cp_y + som->font_height <= d->length);
  text.options = OUTP_T_NONE;
  text.s.s = s;
  text.s.l = strlen (s);
  text.x = som->cp_x;
  text.y = som->cp_y;
  d->class->text_draw (d, &text);
  som->cp_x = 0;
  som->cp_y += som->font_height;
}
    
/* Parses and executes the LIST procedure. */
int
cmd_list (void)
{
  variable casenum_var;
  
  /* (parser) */

  /* Fill in defaults. */
  if (step == NOT_LONG)
    step = 1;
  if (first == NOT_LONG)
    first = 1;
  if (last == NOT_LONG)
    last = LONG_MAX;
  if (!sbc_variables)
    fill_all_vars (&v_variables, &n_variables, FV_NO_SYSTEM | FV_NO_SCRATCH);
  if (n_variables == 0)
    {
      msg (SE, _("No variables specified."));
      return 0;
    }

  /* Verify arguments. */
  if (first > last)
    {
      int t;
      msg (SW, _("The first case (%ld) specified precedes the last case (%ld) "
	   "specified.  The values will be swapped."), first, last);
      t = first;
      first = last;
      last = t;
    }
  if (first < 1)
    {
      msg (SW, _("The first case (%ld) to list is less than 1.  The value is "
	   "being reset to 1."), first);
      first = 1;
    }
  if (last < 1)
    {
      msg (SW, _("The last case (%ld) to list is less than 1.  The value is "
	   "being reset to 1."), last);
      last = 1;
    }
  if (step < 1)
    {
      msg (SW, _("The step value %ld is less than 1.  The value is being "
	   "reset to 1."), step);
      step = 1;
    }

  /* Weighting variable. */
  if (weight == WEIGHT)
    {
      update_weighting (&default_dict);
      if (default_dict.weight_index != -1)
	{
	  int i;

	  for (i = 0; i < n_variables; i++)
	    if (v_variables[i]->index == default_dict.weight_index)
	      break;
	  if (i >= n_variables)
	    {
	      /* Add the weight variable to the end of the variable list. */
	      n_variables++;
	      v_variables = xrealloc (v_variables,
				      n_variables * sizeof (variable *));
	      v_variables[n_variables - 1]
		= default_dict.var[default_dict.weight_index];
	    }
	}
      else
	msg (SW, _("`/FORMAT WEIGHT' specified, but weighting is not on."));
    }

  /* Case number. */
  if (numbering == NUMBERED)
    {
      /* Initialize the case-number variable. */
      strcpy (casenum_var.name, "Case#");
      casenum_var.type = NUMERIC;
      casenum_var.fv = -1;
      casenum_var.print.type = FMT_F;
      casenum_var.print.w = (last == LONG_MAX ? 5 : intlog10 (last));
      casenum_var.print.d = 0;

      /* Add the weight variable at the beginning of the variable list. */
      n_variables++;
      v_variables = xrealloc (v_variables, n_variables * sizeof (variable *));
      memmove (&v_variables[1], &v_variables[0],
	       (n_variables - 1) * sizeof (variable *));
      v_variables[0] = &casenum_var;
    }

#if DEBUGGING
  /* Print out command. */
  debug_print ();
#endif

  determine_layout ();

  case_num = 0;
  procedure (write_all_headers, list_cases, NULL);
  free (line_buf);

  clean_up ();

  return 1;
}

/* Writes headers to all devices.  This is done at the beginning of
   each SPLIT FILE group. */
static void
write_all_headers (void)
{
  outp_driver *d;

  for (d = outp_iterate_enabled_drivers (NULL); d;
       d = outp_iterate_enabled_drivers (d))
    {
      som_driver_ext *som = d->som;

      som->cp_y += som->font_height;		/* Blank line. */
      write_header (d);
    }
}

/* Writes the headers.  Some of them might be vertical; most are
   probably horizontal. */
static void
write_header (outp_driver *d)
{
  list_ext *prc = d->prc;

  if (!prc->header_rows)
    return;
  
  if (n_lines_remaining (d) < prc->header_rows + 1)
    {
      som_internal_eject_page (d);
      assert (n_lines_remaining (d) >= prc->header_rows + 1);
    }

  /* Design the header. */
  if (!prc->header)
    {
      int i, x;

      /* Allocate, initialize header. */
      prc->header = xmalloc (sizeof (char *) * prc->header_rows);
      {
	int w = n_chars_width (d);
	for (i = 0; i < prc->header_rows; i++)
	  {
	    prc->header[i] = xmalloc (w + 1);
	    memset (prc->header[i], ' ', w);
	  }
      }

      /* Put in vertical names. */
      for (i = x = 0; i < prc->n_vertical; i++)
	{
	  variable *v = v_variables[i];
	  int j;

	  memset (&prc->header[prc->header_rows - 1][x], '-', v->print.w);
	  x += v->print.w - 1;
	  for (j = 0; j < (int) strlen (v->name); j++)
	    prc->header[strlen (v->name) - j - 1][x] = v->name[j];
	  x += 2;
	}

      /* Put in horizontal names. */
      for (; i < n_variables; i++)
	{
	  variable *v = v_variables[i];
	  
	  memset (&prc->header[prc->header_rows - 1][x], '-',
		  max (v->print.w, strlen (v->name)));
	  if ((int) strlen (v->name) < v->print.w)
	    x += v->print.w - strlen (v->name);
 	  memcpy (&prc->header[0][x], v->name, strlen (v->name));
	  x += strlen (v->name) + 1;
	}

      /* Add null bytes. */
      for (i = 0; i < prc->header_rows; i++)
	{
	  for (x = n_chars_width (d); x >= 1; x--)
	    if (prc->header[i][x - 1] != ' ')
	      {
		prc->header[i][x] = 0;
		break;
	      }
	  assert (x);
	}
    }

  /* Write out the header, in back-to-front order except for the last line. */
  {
    int i;
    
    for (i = prc->header_rows - 2; i >= 0; i--)
      write_line (d, prc->header[i]);
    write_line (d, prc->header[prc->header_rows - 1]);
  }
}
      
  
/* Frees up all the memory we've allocated. */
static void
clean_up (void)
{
  outp_driver *d;
  
  for (d = outp_iterate_enabled_drivers (NULL); d;
       d = outp_iterate_enabled_drivers (d))
    {
      list_ext *prc = d->prc;
      int i;

      if (prc->header)
	{
	  for (i = 0; i < prc->header_rows; i++)
	    free (prc->header[i]);
	  free (prc->header);
	}
      free (prc);
      
      d->class->text_set_font_by_name (d, "PROP");
    }
  
  free (v_variables);
}

/* Writes string STRING at the current position.  If the text would
   fall off the side of the page, then advance to the next line,
   indenting by amount INDENT. */
static void
write_varname (outp_driver *d, char *string, int indent)
{
  som_driver_ext *som = d->som;

  outp_text text;

  text.options = OUTP_T_NONE;
  text.s.s = string;
  text.s.l = strlen (string);
  d->class->text_metrics (d, &text);
  
  if (som->cp_x + text.h > d->width)
    {
      som->cp_x = indent;
      som->cp_y += som->font_height;
      if (som->cp_y + som->font_height > d->length)
	som_internal_eject_page (d);
    }

  text.x = som->cp_x;
  text.y = som->cp_y;
  d->class->text_draw (d, &text);
  som->cp_x += text.h;
}

/* When we can't fit all the values across the page, we write out all
   the variable names just once.  This is where we do it. */
static void
write_fallback_headers (outp_driver *d)
{
  som_driver_ext *const som = d->som;
  const int max_width = n_chars_width(d) - 10;
  
  int index = 0;
  int width = 0;
  int line_number = 0;

  char *Line = _("Line");
  char *leader = local_alloc (strlen (Line) + INT_DIGITS + 1 + 1);
      
  while (index < n_variables)
    {
      outp_text text;

      /* Ensure that there is enough room for a line of text. */
      if (som->cp_y + som->font_height > d->length)
	som_internal_eject_page (d);
      
      /* The leader is a string like `Line 1: '.  Write the leader. */
      sprintf(leader, "%s %d:", Line, ++line_number);
      text.options = OUTP_T_NONE;
      text.s.s = leader;
      text.s.l = strlen (leader);
      text.x = 0;
      text.y = som->cp_y;
      d->class->text_draw (d, &text);
      som->cp_x = text.h;

      goto entry;
      do
	{
	  width++;

	entry:
	  {
	    int var_width = v_variables[index]->print.w;
	    if (width + var_width > max_width && width != 0)
	      {
		width = 0;
		som->cp_x = 0;
		som->cp_y += som->font_height;
		break;
	      }
	    width += var_width;
	  }
	  
	  {
	    char varname[10];
	    sprintf (varname, " %s", v_variables[index]->name);
	    write_varname (d, varname, text.h);
	  }
	}
      while (++index < n_variables);

    }
  som->cp_x = 0;
  som->cp_y += som->font_height;
  
  local_free (leader);
}

/* There are three possible layouts for the LIST procedure:

   1. If the values and their variables' name fit across the page,
   then they are listed across the page in that way.

   2. If the values can fit across the page, but not the variable
   names, then as many variable names as necessary are printed
   vertically to compensate.

   3. If not even the values can fit across the page, the variable
   names are listed just once, at the beginning, in a compact format,
   and the values are listed with a variable name label at the
   beginning of each line for easier reference.

   This is complicated by the fact that we have to do all this for
   every output driver, not just once.  Oh well, that's the price we
   pay for avoiding the som.  */
static void
determine_layout (void)
{
  outp_driver *d;
  
  /* This is the largest page width of any driver, so we can tell what
     size buffer to allocate. */
  int largest_page_width = 0;
  
  for (d = outp_iterate_enabled_drivers (NULL); d;
       d = outp_iterate_enabled_drivers (d))
    {
      int column;	/* Current column. */
      int width;	/* Accumulated width. */
      int max_width;	/* Page width. */
      
      som_driver_ext *som;
      list_ext *prc;
      
      if (d->som == NULL && !som_init_driver (d))
	continue;
      som = d->som;

      max_width = n_chars_width (d);
      largest_page_width = max (largest_page_width, max_width);

      prc = d->prc = xmalloc (sizeof (list_ext));
      prc->type = 0;
      prc->n_vertical = 0;
      prc->header = NULL;

      /* Try layout #1. */
      for (width = n_variables - 1, column = 0; column < n_variables; column++)
	{
	  variable *v = v_variables[column];
	  width += max (v->print.w, strlen (v->name));
	}
      if (width <= max_width)
	{
	  prc->header_rows = 2;
	  d->class->text_set_font_by_name (d, "FIXED");
	  continue;
	}

      /* Try layout #2. */
      for (width = n_variables - 1, column = 0;
	   column < n_variables && width <= max_width;
	   column++)
	  width += v_variables[column]->print.w;
      
      /* If it fit then we need to determine how many labels can be
         written horizontally. */
      if (width <= max_width)
	{
#ifndef NDEBUG
	  prc->n_vertical = -1;
#endif
	  for (column = n_variables - 1; column >= 0; column--)
	    {
	      variable *v = v_variables[column];
	      int trial_width = (width - v->print.w
				 + max (v->print.w, strlen (v->name)));
	      
	      if (trial_width > max_width)
		{
		  prc->n_vertical = column + 1;
		  break;
		}
	      width = trial_width;
	    }
	  assert(prc->n_vertical != -1);

	  prc->n_vertical = n_variables;
	  /* Finally determine the length of the headers. */
	  for (prc->header_rows = 0, column = 0;
	       column < prc->n_vertical;
	       column++)
	    prc->header_rows = max (prc->header_rows,
				    strlen (v_variables[column]->name));
	  prc->header_rows++;

	  d->class->text_set_font_by_name (d, "FIXED");
	  continue;
	}

      /* Otherwise use the ugly fallback listing format. */
      prc->type = 1;
      prc->header_rows = 0;

      som->cp_y += som->font_height;
      write_fallback_headers (d);
      som->cp_y += som->font_height;
      d->class->text_set_font_by_name (d, "FIXED");
    }

  line_buf = xmalloc (max (256, largest_page_width) + 2);
}

static int
list_cases (ccase *c)
{
  outp_driver *d;
  
  case_num++;
  if (case_num < first || case_num > last
      || (step != 1 && (case_num - first) % step))
    return 1;

  for (d = outp_iterate_enabled_drivers (NULL); d;
       d = outp_iterate_enabled_drivers (d))
    {
      const list_ext *prc = d->prc;
      const int max_width = n_chars_width (d);
      int column;
      int x = 0;

      if (!prc->header_rows)
	x = nsprintf (line_buf, "%8s: ", v_variables[0]->name);
      
      for (column = 0; column < n_variables; column++)
	{
	  variable *v = v_variables[column];
	  int width;

	  if (prc->type == 0 && column >= prc->n_vertical)
	    width = max (strlen (v->name), v->print.w);
	  else
	    width = v->print.w;

	  if (width + x > max_width && x != 0)
	    {
	      if (!n_lines_remaining (d))
		{
		  som_internal_eject_page (d);
		  write_header (d);
		}
	      
	      line_buf[x] = 0;
	      write_line (d, line_buf);

	      x = 0;
	      if (!prc->header_rows)
		x = nsprintf (line_buf, "%8s: ", v->name);
	    }

	  if (width > v->print.w)
	    {
	      memset(&line_buf[x], ' ', width - v->print.w);
	      x += width - v->print.w;
	    }
	  
	  {
	    value value;
	    
	    if (formats[v->print.type].cat & FCAT_STRING)
	      value.c = c->data[v->fv].s;
	    else if (v->fv == -1)
	      value.f = case_num;
	    else
	      value.f = c->data[v->fv].f;
		
	    convert_format_to_string (&line_buf[x], &v->print, &value);
	  }
	  x += v->print.w;
	  
	  line_buf[x++] = ' ';
	}
      
      if (!n_lines_remaining (d))
	{
	  som_internal_eject_page (d);
	  write_header (d);
	}
	      
      line_buf[x] = 0;
      write_line (d, line_buf);
    }
	  
  return 1;
}

/* Debugging output. */

#if DEBUGGING
/* Prints out the command as parsed by cmd_list(). */
static void
debug_print (void)
{
  int i;

  puts ("LIST");
  printf ("  VARIABLES=");
  for (i = 0; i < n_variables; i++)
    {
      if (i)
	putc (' ', stdout);
      fputs (v_variables[i]->name, stdout);
    }

  printf ("\n  /CASES=FROM %ld TO %ld BY %ld\n", first, last, step);

  fputs ("  /FORMAT=", stdout);
  if (numbering == NUMBERED)
    fputs ("NUMBERED", stdout);
  else
    fputs ("UNNUMBERED", stdout);
  putc (' ', stdout);
  if (wrap == WRAP)
    fputs ("WRAP", stdout);
  else
    fputs ("SINGLE", stdout);
  putc (' ', stdout);
  if (weight == WEIGHT)
    fputs ("WEIGHT", stdout);
  else
    fputs ("NOWEIGHT", stdout);
  puts (".");
}
#endif /* DEBUGGING */

/* Local Variables: */
/* mode:c */
/* End: */
