/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#if !som_h
#define som_h 1

/* Structured Output Manager.

   som allows for ASCII output on any system, and high-quality output
   on those systems that support it.  Potential output formats include
   PostScript(R), Windows(TM), and X Windows.  som itself is
   device-independent, turning the device-dependent operations over to
   device drivers.

   som is based on a `recursive tables' model.  The entire output is
   considered to be a single table.  Each cell in the table can
   contain a number or a string, or text, or it can contain another
   table.  They can also contain a `character grid' (which can be used
   to draw an ASCII graph) or a userdraw graphic (which can be used to
   draw a high-res graph).

   A number of dynamic resizing options are available for header cells
   in tables.  Tables can be told to divide themselves into columns
   across the page.  (Only top-level tables may be columnated in this
   fashion.)

   Cells can be undelimited, or they can be separated by spacing or
   single or double lines on any of their four sides.

   Contiguous rectangles of cells in a table can be joined.  After
   they are joined, all the cells together behave as a single cell.

   Memory usage is kept to a minimum by requiring the application to
   have completed a table at the point when it `submits' it to the
   highest-level table that composes an entire output stream.  If the
   driver doesn't need the table after that point, it discards it;
   this is what an ASCII or PostScript driver would do.  The
   application must not modify or examine tables after submitting
   them. */

#include "str.h"
#include "var.h"

/* Table creation options. */
enum
  {
    SOM_CREATE_NONE = 0,	/* No options. */
    SOM_CREATE_ALLOW_RESIZE = 001,	/* Allow resizing of table contents. */
  };

/* Cell content types. */
enum
  {
    SCON_EMPTY,			/* an empty cell */
    SCON_VALUE,			/* a data value */
    SCON_TEXT,			/* rich text */
    SCON_TABLE,			/* subtable */
    SCON_GRID,			/* character grid */
    SCON_DRAW,			/* highres graph */
    SCON_JOIN_AUX,		/* cell other than upper-left in a join */
  };

/* Cell options.  At most one one option from each group may be
   selected. */
enum
  {
    SOPT_NONE = 0,

    SOPT_A_MASK = 00003,	/* alignment mask */
    SOPT_A_RGHT = 00000,	/* right justify */
    SOPT_A_LEFT = 00001,	/* left justify */
    SOPT_A_CNTR = 00002,	/* center */
    SOPT_A_JUST = 00003,	/* fully justify */

    SOPT_X_MASK = 00170,	/* sizing mask */
    SOPT_X_NORM = 00000,	/* ordinary cell */
    SOPT_X_NONE = 00010,	/* does not expand this row/column */
    SOPT_X_HORZ = 00020,	/* expands column as necessary */
    SOPT_X_VERT = 00030,	/* expands row as necessary */
    SOPT_X_BOTH = 00040,	/* expands column if room, row if necessary */
    SOPT_X_JOIN = 00050,	/* upper-left cell in a joined rectangle */
    SOPT_X_HLTL = 00060,	/* expands column if room */
    SOPT_X_SHSP = 00100,	/* aggressively saves horizontal space */
  };

union som_any_cell_union;

/* Empty cell. */
typedef struct
  {
    unsigned char opt;		/* Combination of SOPT_* constants. */
    unsigned char con;		/* One of the SCON_* constants. */
  }
som_empty_cell;

/* String cell. */
typedef struct
  {
    unsigned char opt;		/* Combination of SOPT_* constants. */
    unsigned char con;		/* One of the SCON_* constants. */
    c_string s;			/* Fixed pitch string. */
  }
som_text_cell;

/* Grid cell. */
typedef struct
  {
    unsigned char opt;		/* Combination of SOPT_* constants. */
    unsigned char con;		/* One of the SCON_* constants. */
    int nc, nr;			/* Number of rows, columns. */
    char s[1];			/* Contents (nc*nr bytes). */
  }
som_grid_cell;

/* A cell other than the upper-left cell in a joined rectangle. */
typedef struct
  {
    unsigned char opt;		/* Combination of SOPT_* constants. */
    unsigned char con;		/* One of the SCON_* constants. */
    int c, r;			/* Row, column of som_join_main cell. */
  }
som_join_aux_cell;

/* Table cell. */
typedef struct
  {
    unsigned char opt;		/* Combination of SOPT_* constants. */
    unsigned char con;		/* One of the SCON_* constants. */
    struct som_table_struct *t;	/* Table. */
  }
som_table_cell;

/* Line styles.  These must match output.h:OUTP_L_*. */
enum
  {
    SLIN_0 = 0,			/* No line. */
    SLIN_1THIN = 0,		/* A single line but conserving space. */
    SLIN_1 = 1,			/* Single line. */
    SLIN_2 = 2,			/* Double line. */
    SLIN_3 = 3,			/* Special line of driver-defined style. */
    SLIN_COUNT,			/* Number of line styles. */

    SLIN_SPACING = 0200,	/* Don't draw the line, just reserve space. */
  };

/* A cell. */
typedef union som_any_cell_union
  {
    som_empty_cell any;
    som_text_cell text;
    som_grid_cell grid;
    som_join_aux_cell join_aux;
    som_table_cell table;
  }
som_any_cell;

/* Table options. */
enum
  {
    /* Causes the table to be displayed across the page in multiple
       columns if there is room to do so. */
    STAB_COLUMNATE_MASK = 0x0003,	/* Columnication mask. */
    STAB_COLUMNATE_NONE = 0x0000,	/* Do not columnicate. */
    STAB_COLUMNATE_ACRS = 0x0001,	/* Columnicate columns, across first. */
    STAB_COLUMNATE_DOWN = 0x0002,	/* Columnicate columns, down first. */

    /* Prevents the table from being displayed in multiple horizontal
       chunks.  Mutually exclusive with columnation. */
    STAB_CRUSH = 0x0004,	/* Crush the table. */
  };

/* A table.  Note: rh is of type `unsigned char[nr+1][nc]'; rv is of
   type `unsigned char[nr][nc+1]'. */
typedef struct som_table_struct
  {
    /* Parent info. */
    struct som_table_struct *owner;
    /* Parent table. */
    int x, y;			/* Column, row in parent table. */

    /* Table info. */
    int flags;			/* Table options. */
    int nc, nr;			/* Number of columns, rows. */
    int l, r, t, b;		/* Number of header rows on each side. */
    int group;			/* When columnating, # of rows per group. */
    unsigned char *rh, *rv;	/* Types of horizontal/vertical rules. */
    som_any_cell **c;		/* Contents. */

    /* Calculated at output time. */
    int *w, *h;			/* Width/height of columns/rows/rules. */
  }
som_table;

/* A header that is tacked onto a cell that is the upper-left corner
   of a joined rectangle.  This header precedes the cell contents
   proper.  The cell pointer points to the cell contents, not to the
   header. */
typedef union
  {
    char pad[16];		/* Ensures proper alignment for every
				   machine I know of. */
    struct
      {
	int nc;			/* Number of joined columns. */
	int nr;			/* Number of joined rows. */
      }
    d;
  }
som_join_header;

/* Table submission types. */
enum
  {
    SOM_SUB_ORDINARY,		/* Just an ordinary table submission. */
    SOM_SUB_COLUMNATE_DOWN,	/* Divide the table into columns. */
  };

/* Table submission options. */
enum
  {
    SOM_TOPT_NONE = 0,
    SOM_TOPT_HEADER = 0001,	/* Omit the header line. */
    SOM_TOPT_SPACING = 0002,	/* Omit the spacing before the table. */
    SOM_TOPT_CRUSH = 0004,	/* Don't break the table horizontally. */

    /* One view of the partial table options. */
    SOM_TOPT_PARTIAL_MASK = 0060,	/* Partial tables (crushed tables only). */
    SOM_TOPT_PARTIAL_NONE = 0000,	/* Not a partial table. */
    SOM_TOPT_PARTIAL_BEG = 0020,	/* The first part of a partial table. */
    SOM_TOPT_PARTIAL_MID = 0060,	/* A middle part of a partial table. */
    SOM_TOPT_PARTIAL_END = 0040,	/* The final part of a partial table. */

    /* Another view of the partial table options. */
    SOM_TOPT_PARTIAL_OMIT_BTM = 0020,	/* Don't draw the bottom headers. */
    SOM_TOPT_PARTIAL_OMIT_TOP = 0040,	/* Don't draw the top headers. */
  };

/* `Simply submit these 7 forms in triplicate...' */
typedef struct
  {
    int type;			/* Submission type. */
    int options;		/* SOM_TOPT_*. */
    som_table *t;		/* The table being submitted. */
    a_string title;		/* Table title. */
  }
som_submission_form;

/* Tables. */
void som_push_workspace (void);
void som_pop_workspace (void);
void som_check_workspace (void);
som_table *som_create_table (int nc, int nr, int create_flags);
void som_set_table_height (som_table *, int nr);
void som_change_table_size (som_table * t, int nc, int nr);

/* Rules. */
void som_hline (som_table *, int style, int x1, int x2, int y);
void som_vline (som_table *, int style, int x, int y1, int y2);
void som_box (som_table *, int f_h, int f_v, int i_h, int i_v,
	      int x1, int y1, int x2, int y2);

/* Properties of entire tables. */
void som_set_headers (som_table *, int l, int r, int t, int b);
void som_columnate (som_table *, int across, int group);

/* Not implemented. */
void som_join_cells (som_table *, int x1, int y1, int x2, int y2);
void som_insert_table (som_table * parent, int c, int r, som_table * child,
		       unsigned char opt);

/* Submission. */
void som_new_series (void);
void som_submit_table (som_submission_form * f);
void som_eject_page (void);

/* Options for passing to som_set_text() and som_output_text().  May
   be OR'd with the SOPT_* cell attributes and passed in `opt'. */
enum
  {
    SOT_NONE = 0,		/* No options. */
    SOT_RICH = 0x0100,		/* The text string is rich text. */
    /*   ^^^^^^^ FIXME: Doesn't work since \0's can't be put in strings. */
    SOT_PRINTF = 0x0200,	/* Pass the text string through sprintf. */
    SOT_TITLE = 0x0500,		/* Apply title attributes (implies SOT_RICH). */
    SOT_FIX = 0x1000,		/* Apply fixed font (exclusive of SOT_RICH). */
  };

/* Cells. */
void som_set_value (som_table *, int c, int r, unsigned char opt,
		    value * v, fmt_spec * f, int count);
void som_set_float (som_table *, int c, int r, unsigned char opt,
		    double v, int ndec);

extern char zero_length[1];

void som_set_text (som_table *, int c, int r, unsigned opt, char *, ...)
     __attribute__ ((format (printf, 5, 6)));
  
void som_output_text (int options, char *string, ...)
     __attribute__ ((format (printf, 2, 3)));
  
#define blank_line() som_output_text (SOPT_NONE, zero_length)

void som_set_null (som_table *, int c, int r);

#endif /* som_h */
